# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations

# Custom Library

# Custom Packages
import AthenaLib._info.formatting as f
from AthenaLib._info._v import _version

# ----------------------------------------------------------------------------------------------------------------------
# - Code -
# ----------------------------------------------------------------------------------------------------------------------
def info(*, to_str: bool = False) -> None | str:
    line = "-" * 128
    header = f.header(f"""{line}
{f.title("AthenaLib", to_str)} v{_version()}
is made by Andreas Sas and is a standard library for "{f.reference("Athena...", to_str)}" packages
{line}
""", to_str)

    body = f"""
Package setup:
    {f.sub_modules("decorators", to_str)} : Decorators for functions or methods

    {f.sub_modules("fixes", to_str)} : Functions which resolve some issues which are brought up with Python
        An Example of this is nested asyncio loops. (A third party package is used for this, "{f.reference('nest_asyncio', to_str)}")

    {f.sub_modules("models", to_str)} : Basic classes which are meant to hold data. 
        These sort of classes are not meant to house a lot methods which interact with lots of other classes

    {f.sub_modules("functions", to_str)} : A collection of basic functions to handle various tasks
        This is often split up further into finer groups, like file manipulation, sorting algorithms, etc...
"""

    text = f"{header}{body}{line}"

    # export to console or string
    if to_str:
        return text
    else:
        print(text)


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
from AthenaLib.models.bezier import CubicBezier
from AthenaLib.models.degree import Degree
from AthenaLib.models.length_absolute import (
    AbsoluteLength, Pixel, Pica, Point, Inch, Meter, DeciMeter, CentiMeter, MilliMeter
)
from AthenaLib.models.length_relative import (
    RelativeLength,ElementFontSize,ElementFontHeight,ZeroCharacterWidth,RootElementFontSize,ViewportWidthPercent,
    ViewportHeightPercent,ViewportLargerPercent,ViewportSmallerPercent
)
from AthenaLib.models.percent import Percent
from AthenaLib.models.time import (
    Minute,Second,MilliSecond,Hour
)
from AthenaLib.models.url import Url
from AthenaLib.models.vectors import Vector1D, Vector2D, Vector3D
from AthenaLib.models.version import Version


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from typing import Callable

# Custom Library

# Custom Packages
import AthenaLib.functions.dunder_math as dunder_math

# ----------------------------------------------------------------------------------------------------------------------
# - All -
# ----------------------------------------------------------------------------------------------------------------------
__all__=[
    "CubicBezier"
]

# ----------------------------------------------------------------------------------------------------------------------
# - Support Functions -
# ----------------------------------------------------------------------------------------------------------------------
def dunder_func(func:Callable,left:CubicBezier,right:CubicBezier|int|float|tuple):
    if type(left) is type(right):
        return func(
            (left.x1,left.y1,left.x2,left.y2),
            (right.x1,right.y1,right.x2,right.y2)
        )
    elif isinstance(right, (int, float)):
        return func(
            (left.x1,left.y1,left.x2,left.y2),
            (right,right,right,right)
        )
    elif isinstance(right, tuple) and len(right) == 4:
        return func(
            (left.x1,left.y1,left.x2,left.y2),
            right
        )
    else:
        return NotImplemented
# ----------------------------------------------------------------------------------------------------------------------
# - Classes -
# ----------------------------------------------------------------------------------------------------------------------
class CubicBezier:
    _x1:float
    _y1:float
    _x2:float
    _y2:float

    def __init__(self, x1:float,y1:float,x2:float,y2:float):
        self.x1, self.y1, self.x2, self.y2 = x1,y1,x2,y2

    @property
    def x1(self) -> float:
        return self._x1
    @x1.setter
    def x1(self, value):
        if not isinstance(value, int|float):
            raise TypeError
        self._x1 = max(min(value, 1.0),0.0)

    @property
    def y1(self) -> float:
        return self._y1
    @y1.setter
    def y1(self, value):
        if not isinstance(value, int|float):
            raise TypeError
        self._y1 = value

    @property
    def x2(self) -> float:
        return self._x2
    @x2.setter
    def x2(self, value):
        if not isinstance(value, int|float):
            raise TypeError
        self._x2 = max(min(value, 1.0),0.0)

    @property
    def y2(self) -> float:
        return self._y2
    @y2.setter
    def y2(self, value):
        if not isinstance(value, int|float):
            raise TypeError
        self._y2 = value

    # ------------------------------------------------------------------------------------------------------------------
    # - cast dunders -
    # ------------------------------------------------------------------------------------------------------------------
    def __abs__(self) -> CubicBezier:
        return CubicBezier(abs(self.x1),abs(self.y1),abs(self.x2),abs(self.y2))
    def __hash__(self) -> int:
        return hash((self.x1,self.y1,self.x2,self.y2))
    def __repr__(self) -> str:
        return f"CubicBezier(x1={self.x1},y1={self.y1},x2={self.x2},y2={self.y2})"
    def __round__(self, n=None):
        return type(self)(*(round(i, n) for i in (self.x1, self.y1, self.x2, self.y2)))

    # ------------------------------------------------------------------------------------------------------------------
    # - Comparison Operations -
    # ------------------------------------------------------------------------------------------------------------------
    def __eq__(self, other: CubicBezier | int | float) -> bool:
        return dunder_func(func=dunder_math.eq, left=self, right=other)
    def __ne__(self, other: CubicBezier | int | float) -> bool:
        return dunder_func(func=dunder_math.ne, left=self, right=other)
    def __gt__(self, other: CubicBezier | int | float) -> bool:
        return dunder_func(func=dunder_math.gt, left=self, right=other)
    def __lt__(self, other: CubicBezier | int | float) -> bool:
        return dunder_func(func=dunder_math.lt, left=self, right=other)
    def __ge__(self, other: CubicBezier | int | float) -> bool:
        return dunder_func(func=dunder_math.ge, left=self, right=other)
    def __le__(self, other: CubicBezier | int | float) -> bool:
        return dunder_func(func=dunder_math.le, left=self, right=other)

    # ------------------------------------------------------------------------------------------------------------------
    # - math Operations -
    # ------------------------------------------------------------------------------------------------------------------
    def __add__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.add, left=self, right=other)) is NotImplemented:
            return result
        return type(self)(*result)
    def __sub__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.sub, left=self, right=other)) is NotImplemented:
            return result
        return type(self)(*result)
    def __mul__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.mul, left=self, right=other)) is NotImplemented:
            return result
        return type(self)(*result)
    def __floordiv__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.floordiv, left=self, right=other)) is NotImplemented:
            return result
        return type(self)(*result)
    def __truediv__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.truediv, left=self, right=other)) is NotImplemented:
            return result
        return type(self)(*result)
    def __pow__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.power, left=self, right=other)) is NotImplemented:
            return result
        return type(self)(*result)
    def __mod__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.mod, left=self, right=other)) is NotImplemented:
            return result
        return type(self)(*result)

    def __iadd__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.add, left=self, right=other)) is NotImplemented:
            return result
        self.x1,self.y1,self.x2,self.y2 = result
        return self
    def __isub__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.sub, left=self, right=other)) is NotImplemented:
            return result
        self.x1,self.y1,self.x2,self.y2 = result
        return self
    def __imul__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.mul, left=self, right=other)) is NotImplemented:
            return result
        self.x1,self.y1,self.x2,self.y2 = result
        return self
    def __ifloordiv__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.floordiv, left=self, right=other)) is NotImplemented:
            return result
        self.x1,self.y1,self.x2,self.y2 = result
        return self
    def __itruediv__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.truediv, left=self, right=other)) is NotImplemented:
            return result
        self.x1,self.y1,self.x2,self.y2 = result
        return self
    def __ipow__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.power, left=self, right=other)) is NotImplemented:
            return result
        self.x1,self.y1,self.x2,self.y2 = result
        return self
    def __imod__(self, other) -> CubicBezier:
        if (result:= dunder_func(func=dunder_math.mod, left=self, right=other)) is NotImplemented:
            return result
        self.x1,self.y1,self.x2,self.y2 = result
        return self


# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from abc import ABC, abstractmethod

# Custom Library

# Custom Packages

# ----------------------------------------------------------------------------------------------------------------------
# - All -
# ----------------------------------------------------------------------------------------------------------------------
__all__ = [
    "RelativeLength","ElementFontSize", "ElementFontHeight", "ZeroCharacterWidth", "RootElementFontSize",
    "ViewportWidthPercent", "ViewportHeightPercent", "ViewportLargerPercent", "ViewportSmallerPercent"
]

# ----------------------------------------------------------------------------------------------------------------------
# - Support Functions -
# ----------------------------------------------------------------------------------------------------------------------
def _relativeLengthConversionInput(fnc):
    def wrapper(self:RelativeLength, *args, **kwargs):
        if isinstance(args, tuple):
            other,*args_ = args
        else:
            other = args
            args_ = ()

        if isinstance(other, int|float):
            return fnc(self, type(self)(value=other), *args_, **kwargs)
        elif isinstance(other, type(self)):
            return fnc(self, *args, **kwargs)
        else:
            return NotImplemented
    return wrapper

# ----------------------------------------------------------------------------------------------------------------------
# - Classes -
# ----------------------------------------------------------------------------------------------------------------------
class RelativeLength(ABC):
    _value:int|float
    def __init__(self, value: int | float | RelativeLength):
        self.value = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not isinstance(value, int|float):
            raise TypeError
        self._value = value

    # ------------------------------------------------------------------------------------------------------------------
    # - cast dunders -
    # ------------------------------------------------------------------------------------------------------------------
    def __int__(self) -> int:
        return int(self.value)
    def __float__(self) -> float:
        return float(self.value)
    def __abs__(self) -> RelativeLength:
        return type(self)(abs(self.value))
    def __round__(self, n=None):
        return type(self)(round(self.value, n))
    def __hash__(self) -> int:
        return hash(self.value)

    @abstractmethod
    def __repr__(self) -> str:...

    # ------------------------------------------------------------------------------------------------------------------
    # - Comparison Operations -
    # ------------------------------------------------------------------------------------------------------------------
    @_relativeLengthConversionInput
    def __eq__(self, other: RelativeLength | int | float) -> bool:
        return self.value == other.value
    @_relativeLengthConversionInput
    def __ne__(self, other: RelativeLength | int | float) -> bool:
        return self.value != other.value
    @_relativeLengthConversionInput
    def __gt__(self, other: RelativeLength | int | float) -> bool:
        return self.value > other.value
    @_relativeLengthConversionInput
    def __lt__(self, other: RelativeLength | int | float) -> bool:
        return self.value < other.value
    @_relativeLengthConversionInput
    def __ge__(self, other: RelativeLength | int | float) -> bool:
        return self.value >= other.value
    @_relativeLengthConversionInput
    def __le__(self, other: RelativeLength | int | float) -> bool:
        return self.value <= other.value

    # ------------------------------------------------------------------------------------------------------------------
    # - math Operations -
    # ------------------------------------------------------------------------------------------------------------------
    @_relativeLengthConversionInput
    def __add__(self, other: RelativeLength | int | float) -> RelativeLength:
        return type(self)(self.value + other.value)
    @_relativeLengthConversionInput
    def __sub__(self, other: RelativeLength | int | float) -> RelativeLength:
        return type(self)(self.value - other.value)
    @_relativeLengthConversionInput
    def __mul__(self, other: RelativeLength | int | float) -> RelativeLength:
        return type(self)(self.value * other.value)
    @_relativeLengthConversionInput
    def __floordiv__(self, other: RelativeLength | int | float) -> RelativeLength:
        return type(self)(self.value // other.value)
    @_relativeLengthConversionInput
    def __truediv__(self, other: RelativeLength | int | float) -> RelativeLength:
        return type(self)(self.value / other.value)
    @_relativeLengthConversionInput
    def __pow__(self, other: RelativeLength | int | float) -> RelativeLength:
        return type(self)(self.value ** other.value)
    @_relativeLengthConversionInput
    def __mod__(self, other: RelativeLength | int | float) -> RelativeLength:
        return type(self)(self.value % other.value)

    @_relativeLengthConversionInput
    def __iadd__(self, other: RelativeLength | int | float):
        self.value += other.value
        return self
    @_relativeLengthConversionInput
    def __isub__(self, other: RelativeLength | int | float):
        self.value -= other.value
        return self
    @_relativeLengthConversionInput
    def __imul__(self, other: RelativeLength | int | float):
        self.value *= other.value
        return self
    @_relativeLengthConversionInput
    def __ifloordiv__(self, other: RelativeLength | int | float):
        self.value //= other.value
        return self
    @_relativeLengthConversionInput
    def __itruediv__(self, other: RelativeLength | int | float):
        self.value /= other.value
        return self
    @_relativeLengthConversionInput
    def __ipow__(self, other: RelativeLength | int | float):
        self.value **= other.value
        return self
    @_relativeLengthConversionInput
    def __imod__(self, other: RelativeLength | int | float):
        self.value %= other.value
        return self

# ----------------------------------------------------------------------------------------------------------------------
class ElementFontSize(RelativeLength):
    def __repr__(self) -> str:
        return f"ElementFontSize(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class ElementFontHeight(RelativeLength):
    def __repr__(self) -> str:
        return f"ElementFontHeight(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class ZeroCharacterWidth(RelativeLength):
    def __repr__(self) -> str:
        return f"ZeroCharacterWidth(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class RootElementFontSize(RelativeLength):
    def __repr__(self) -> str:
        return f"RootElementFontSize(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class ViewportWidthPercent(RelativeLength):
    def __repr__(self) -> str:
        return f"ViewportWidthPercent(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class ViewportHeightPercent(RelativeLength):
    def __repr__(self) -> str:
        return f"ViewportHeightPercent(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class ViewportSmallerPercent(RelativeLength):
    def __repr__(self) -> str:
        return f"ViewportSmallerPercent(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class ViewportLargerPercent(RelativeLength):
    def __repr__(self) -> str:
        return f"ViewportLargerPercent(value={self.value})"




# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from abc import ABC, abstractmethod

# Custom Library

# Custom Packages

# ----------------------------------------------------------------------------------------------------------------------
# - All -
# ----------------------------------------------------------------------------------------------------------------------
__all__=[
    "absoluteLengthConversion", "AbsoluteLength",
    "Pixel", "Pica", "Point",
    "Inch",
    "Meter", "DeciMeter", "CentiMeter", "MilliMeter"
]

# ----------------------------------------------------------------------------------------------------------------------
# - Support Functions -
# ----------------------------------------------------------------------------------------------------------------------
def absoluteLengthConversion(original: int | float | AbsoluteLength, cast: int | float | AbsoluteLength, *, pixels_per_inch:int=96) -> int | float | AbsoluteLength:
    match original, cast:
        # General catches for throwing to and from numbers
        case int()|float(), AbsoluteLength():
            return type(cast)(original)
        case AbsoluteLength(), int() | float():
            return type(cast)(original.value)

        # To length Types, if both types are the same (Pixel->Pixel)
        case a,b if type(a) is type(b) and isinstance(a, AbsoluteLength) and isinstance(b, AbsoluteLength):
            return type(cast)(original.value)

        # Pica
        case Pica(), Pixel():
            return Pixel(original.value * 16)
        case Pica(), Point():
            return Point(original.value * 12)
        case Pica(), Inch():
            return Inch(original.value/6)
        case Pica(), MilliMeter():
            inch_value = original.value/6
            return MilliMeter(inch_value*25.4)
        case Pica(), CentiMeter():
            inch_value = original.value/6
            return CentiMeter(inch_value*2.54)
        case Pica(), DeciMeter():
            inch_value = original.value/6
            return DeciMeter(inch_value*.254)
        case Pica(), Meter():
            inch_value = original.value/6
            return Meter(inch_value*.0254)

        # Point
        case Point(), Pixel():
            inch_value = original.value/72
            return Pixel(inch_value*pixels_per_inch)
        case Point(), Pica():
            return Pica(original.value/12)
        case Point(), Inch():
            return Inch(original.value/72)
        case Point(), MilliMeter():
            inch_value = original.value/72
            return MilliMeter(inch_value*25.4)
        case Point(), CentiMeter():
            inch_value = original.value/72
            return CentiMeter(inch_value*2.54)
        case Point(), DeciMeter():
            inch_value = original.value/72
            return DeciMeter(inch_value*.254)
        case Point(), Meter():
            inch_value = original.value/72
            return Meter(inch_value*.0254)

        # Pixel
        case Pixel(), Pica():
            return Pica(original.value/16)
        case Pixel(), Point():
            inch_value = original.value/pixels_per_inch
            return Point(inch_value*72)
        case Pixel(), Inch():
            return Inch(original.value/pixels_per_inch)
        case Pixel(), MilliMeter():
            inch_value = original.value/pixels_per_inch
            return MilliMeter(inch_value*25.4)
        case Pixel(), CentiMeter():
            inch_value = original.value/pixels_per_inch
            return CentiMeter(inch_value*2.54)
        case Pixel(), DeciMeter():
            inch_value = original.value/pixels_per_inch
            return DeciMeter(inch_value*.254)
        case Pixel(), Meter():
            inch_value = original.value/pixels_per_inch
            return Meter(inch_value*.0254)

        # Inch
        case Inch(), Pixel():
            return Pixel(original.value*pixels_per_inch)
        case Inch(), Pica():
            return Pica(original.value/0.166666667)
        case Inch(), Point():
            return Point(original.value*72)
        case Inch(), MilliMeter():
            return MilliMeter(original.value*25.4)
        case Inch(), CentiMeter():
            return CentiMeter(original.value*2.54)
        case Inch(), DeciMeter():
            return DeciMeter(original.value*.254)
        case Inch(), Meter():
            return Meter(original.value*.0254)

        # MilliMeter
        case MilliMeter(), Pixel():
            inch_value = original.value/25.4
            return Pixel(inch_value*96)
        case MilliMeter(), Pica():
            inch_value = original.value/25.4
            return Pica(inch_value*6)
        case MilliMeter(), Point():
            inch_value = original.value/25.4
            return Point(72*inch_value)
        case MilliMeter(), Inch():
            return Inch(original.value/25.4)
        case MilliMeter(), CentiMeter():
            return CentiMeter(original.value/10)
        case MilliMeter(), DeciMeter():
            return DeciMeter(original.value/100)
        case MilliMeter(), Meter():
            return Meter(original.value/1000)

        # CentiMeter
        case CentiMeter(), Pixel():
            inch_value = original.value/2.54
            return Pixel(inch_value*96)
        case CentiMeter(), Pica():
            inch_value = original.value/2.54
            return Pica(inch_value*6)
        case CentiMeter(), Point():
            inch_value = original.value/2.54
            return Point(72*inch_value)
        case CentiMeter(), Inch():
            return Inch(original.value/2.54)
        case CentiMeter(), MilliMeter():
            return MilliMeter(original.value*10)
        case CentiMeter(), DeciMeter():
            return DeciMeter(original.value/10)
        case CentiMeter(), Meter():
            return Meter(original.value/100)

        # DeciMeter
        case DeciMeter(), Pixel():
            inch_value = original.value/.254
            return Pixel(inch_value*96)
        case DeciMeter(), Pica():
            inch_value = original.value/.254
            return Pica(inch_value*6)
        case DeciMeter(), Point():
            inch_value = original.value/.254
            return Point(72*inch_value)
        case DeciMeter(), Inch():
            return Inch(original.value/.254)
        case DeciMeter(), MilliMeter():
            return MilliMeter(original.value*100)
        case DeciMeter(), CentiMeter():
            return CentiMeter(original.value*10)
        case DeciMeter(), Meter():
            return Meter(original.value/10)

        # Meter
        case Meter(), Pixel():
            inch_value = original.value/.0254
            return Pixel(inch_value*96)
        case Meter(), Pica():
            inch_value = original.value/.0254
            return Pica(inch_value*6)
        case Meter(), Point():
            inch_value = original.value/.0254
            return Point(72*inch_value)
        case Meter(), Inch():
            return Inch(original.value/.0254)
        case Meter(), MilliMeter():
            return MilliMeter(original.value*1000)
        case Meter(), CentiMeter():
            return CentiMeter(original.value*100)
        case Meter(), DeciMeter():
            return DeciMeter(original.value*10)

        case _:
            return NotImplemented

def _absoluteLengthConversionInput(fnc):
    def wrapper(self:AbsoluteLength, *args, **kwargs):
        if isinstance(args, tuple):
            other,*_ = args
        else:
            other, _ = args, ()

        if (otherConverted := absoluteLengthConversion(other, self, pixels_per_inch=self.pixels_per_inch)) is NotImplemented:
            return NotImplemented
        return fnc(self,otherConverted, *_,**kwargs)
    return wrapper

# ----------------------------------------------------------------------------------------------------------------------
# - Classes -
# ----------------------------------------------------------------------------------------------------------------------
class AbsoluteLength(ABC):
    _value:int|float
    pixels_per_inch:int
    def __init__(self, value: int | float | AbsoluteLength):
        self.value = value
        self.pixels_per_inch=96

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not isinstance(value, int|float):
            raise TypeError
        self._value = value

    # ------------------------------------------------------------------------------------------------------------------
    # - cast dunders -
    # ------------------------------------------------------------------------------------------------------------------
    def __int__(self) -> int:
        return int(self.value)
    def __float__(self) -> float:
        return float(self.value)
    def __abs__(self) -> AbsoluteLength:
        return  type(self)(abs(self.value))
    def __round__(self, n=None):
        return  type(self)(round(self.value, n))
    def __hash__(self) -> int:
        return hash(self.value)

    @abstractmethod
    def __repr__(self) -> str:...

    # ------------------------------------------------------------------------------------------------------------------
    # - Comparison Operations -
    # ------------------------------------------------------------------------------------------------------------------
    @_absoluteLengthConversionInput
    def __eq__(self, other: AbsoluteLength | int | float) -> bool:
        return self.value == other.value
    @_absoluteLengthConversionInput
    def __ne__(self, other: AbsoluteLength | int | float) -> bool:
        return self.value != other.value
    @_absoluteLengthConversionInput
    def __gt__(self, other: AbsoluteLength | int | float) -> bool:
        return self.value > other.value
    @_absoluteLengthConversionInput
    def __lt__(self, other: AbsoluteLength | int | float) -> bool:
        return self.value < other.value
    @_absoluteLengthConversionInput
    def __ge__(self, other: AbsoluteLength | int | float) -> bool:
        return self.value >= other.value
    @_absoluteLengthConversionInput
    def __le__(self, other: AbsoluteLength | int | float) -> bool:
        return self.value <= other.value

    # ------------------------------------------------------------------------------------------------------------------
    # - math Operations -
    # ------------------------------------------------------------------------------------------------------------------
    @_absoluteLengthConversionInput
    def __add__(self, other: AbsoluteLength | int | float) -> AbsoluteLength:
        return type(self)(self.value + other.value)
    @_absoluteLengthConversionInput
    def __sub__(self, other: AbsoluteLength | int | float) -> AbsoluteLength:
        return type(self)(self.value - other.value)
    @_absoluteLengthConversionInput
    def __mul__(self, other: AbsoluteLength | int | float) -> AbsoluteLength:
        return type(self)(self.value * other.value)
    @_absoluteLengthConversionInput
    def __floordiv__(self, other: AbsoluteLength | int | float) -> AbsoluteLength:
        return type(self)(self.value // other.value)
    @_absoluteLengthConversionInput
    def __truediv__(self, other: AbsoluteLength | int | float) -> AbsoluteLength:
        return type(self)(self.value / other.value)
    @_absoluteLengthConversionInput
    def __pow__(self, other: AbsoluteLength | int | float) -> AbsoluteLength:
        return type(self)(self.value ** other.value)
    @_absoluteLengthConversionInput
    def __mod__(self, other: AbsoluteLength | int | float) -> AbsoluteLength:
        return type(self)(self.value % other.value)

    @_absoluteLengthConversionInput
    def __iadd__(self, other: AbsoluteLength | int | float):
        self.value += other.value
        return self
    @_absoluteLengthConversionInput
    def __isub__(self, other: AbsoluteLength | int | float):
        self.value -= other.value
        return self
    @_absoluteLengthConversionInput
    def __imul__(self, other: AbsoluteLength | int | float):
        self.value *= other.value
        return self
    @_absoluteLengthConversionInput
    def __ifloordiv__(self, other: AbsoluteLength | int | float):
        self.value //= other.value
        return self
    @_absoluteLengthConversionInput
    def __itruediv__(self, other: AbsoluteLength | int | float):
        self.value /= other.value
        return self
    @_absoluteLengthConversionInput
    def __ipow__(self, other: AbsoluteLength | int | float):
        self.value **= other.value
        return self
    @_absoluteLengthConversionInput
    def __imod__(self, other: AbsoluteLength | int | float):
        self.value %= other.value
        return self

# ----------------------------------------------------------------------------------------------------------------------
class Pica(AbsoluteLength):
    def __repr__(self) -> str:
        return f"Pica(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class Point(AbsoluteLength):
    def __repr__(self) -> str:
        return f"Point(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class Pixel(AbsoluteLength):
    def __repr__(self) -> str:
        return f"Pixel(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class Inch(AbsoluteLength):
    def __repr__(self) -> str:
        return f"Inch(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class MilliMeter(AbsoluteLength):
    def __repr__(self) -> str:
        return f"MilliMeter(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class CentiMeter(AbsoluteLength):
    def __repr__(self) -> str:
        return f"CentiMeter(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class DeciMeter(AbsoluteLength):
    def __repr__(self) -> str:
        return f"DeciMeter(value={self.value})"

# ----------------------------------------------------------------------------------------------------------------------
class Meter(AbsoluteLength):
    def __repr__(self) -> str:
        return f"Meter(value={self.value})"



# ----------------------------------------------------------------------------------------------------------------------
# - Package Imports -
# ----------------------------------------------------------------------------------------------------------------------
# General Packages
from __future__ import annotations
from abc import ABC, abstractmethod

# Custom Library

# Custom Packages

# ----------------------------------------------------------------------------------------------------------------------
# - All -
# ----------------------------------------------------------------------------------------------------------------------
__all__=[
    "Minute", "Second", "MilliSecond", "Hour",
    "timeConversion"
]

# ----------------------------------------------------------------------------------------------------------------------
# - Support Functions -
# ----------------------------------------------------------------------------------------------------------------------
def timeConversion(original: int | float | TimeValue, cast: int | float | TimeValue) -> int | float | TimeValue:
    match original, cast:
        # General catches for throwing to and from numbers
        case int()|float(), TimeValue():
            return type(cast)(original)
        case TimeValue(), int() | float():
            return type(cast)(original.value)

        # To Time Types, if both types are the same (Second-> Second, Mili-> Mili)
        case a,b if type(a) is type(b):
            return type(cast)(original.value)

        # Conversion functions between types
        case Second(), MilliSecond():
            return MilliSecond(original.value * 1000)
        case Second(), Minute():
            return Minute(original.value/60)
        case Second(), Hour():
            return Hour(original.value/60/60)

        case MilliSecond(), Second():
            return Second(original.value/1000)
        case MilliSecond(), Minute():
            return Minute(original.value/1000/60)
        case MilliSecond(), Hour():
            return Hour(original.value/1000/60/60)

        case Minute(), MilliSecond():
            return MilliSecond(original.value*60*1000)
        case Minute(), Second():
            return Second(original.value*60)
        case Minute(), Hour():
            return Hour(original.value/60)

        case Hour(), MilliSecond():
            return MilliSecond(original.value*60*1000*60)
        case Hour(), Second():
            return Second(original.value*60*60)
        case Hour(), Minute():
            return Minute(original.value*60)

        case _:
            return NotImplemented

def _timeConversionInput(fnc):
    def wrapper(self,*args,**kwargs):
        if isinstance(args, tuple):
            other,*_ = args
        else:
            other, _ = args, ()

        if (otherConverted := timeConversion(other, self)) is NotImplemented:
            return NotImplemented
        return fnc(self,otherConverted, *_,**kwargs)
    return wrapper

# ----------------------------------------------------------------------------------------------------------------------
# - Classes -
# ----------------------------------------------------------------------------------------------------------------------
class TimeValue(ABC):
    _value:int|float

    def __init__(self, value: int | float | TimeValue):
        self.value = value

    def __hash__(self) -> int:
        return hash(self.value)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if not isinstance(value, int|float):
            raise TypeError
        self._value = value

    # ------------------------------------------------------------------------------------------------------------------
    # - cast dunders -
    # ------------------------------------------------------------------------------------------------------------------
    def __int__(self) -> int:
        return int(self.value)
    def __float__(self) -> float:
        return float(self.value)
    def __abs__(self) -> TimeValue:
        return type(self)(abs(self.value))
    def __round__(self, n=None):
        return type(self)(round(self.value, n))

    @abstractmethod
    def __repr__(self) -> str:...

    # ------------------------------------------------------------------------------------------------------------------
    # - Comparison Operations -
    # ------------------------------------------------------------------------------------------------------------------
    @_timeConversionInput
    def __eq__(self, other: TimeValue | int | float) -> bool:
        return self.value == other.value
    @_timeConversionInput
    def __ne__(self, other: TimeValue | int | float) -> bool:
        return self.value != other.value
    @_timeConversionInput
    def __gt__(self, other: TimeValue | int | float) -> bool:
        return self.value > other.value
    @_timeConversionInput
    def __lt__(self, other: TimeValue | int | float) -> bool:
        return self.value < other.value
    @_timeConversionInput
    def __ge__(self, other: TimeValue | int | float) -> bool:
        return self.value >= other.value
    @_timeConversionInput
    def __le__(self, other: TimeValue | int | float) -> bool:
        return self.value <= other.value

    # ------------------------------------------------------------------------------------------------------------------
    # - math Operations -
    # ------------------------------------------------------------------------------------------------------------------
    @_timeConversionInput
    def __add__(self, other: TimeValue | int | float) -> TimeValue:
        return type(self)(self.value + other.value)
    @_timeConversionInput
    def __sub__(self, other: TimeValue | int | float) -> TimeValue:
        return type(self)(self.value - other.value)
    @_timeConversionInput
    def __mul__(self, other: TimeValue | int | float) -> TimeValue:
        return type(self)(self.value * other.value)
    @_timeConversionInput
    def __floordiv__(self, other: TimeValue | int | float) -> TimeValue:
        return type(self)(self.value // other.value)
    @_timeConversionInput
    def __truediv__(self, other: TimeValue | int | float) -> TimeValue:
        return type(self)(self.value / other.value)
    @_timeConversionInput
    def __pow__(self, other: TimeValue | int | float) -> TimeValue:
        return type(self)(self.value ** other.value)
    @_timeConversionInput
    def __mod__(self, other: TimeValue | int | float) -> TimeValue:
        return type(self)(self.value % other.value)

    @_timeConversionInput
    def __iadd__(self, other: TimeValue | int | float):
        self.value += other.value
        return self
    @_timeConversionInput
    def __isub__(self, other: TimeValue | int | float):
        self.value -= other.value
        return self
    @_timeConversionInput
    def __imul__(self, other: TimeValue | int | float):
        self.value *= other.value
        return self
    @_timeConversionInput
    def __ifloordiv__(self, other: TimeValue | int | float):
        self.value //= other.value
        return self
    @_timeConversionInput
    def __itruediv__(self, other: TimeValue | int | float):
        self.value /= other.value
        return self
    @_timeConversionInput
    def __ipow__(self, other: TimeValue | int | float):
        self.value **= other.value
        return self
    @_timeConversionInput
    def __imod__(self, other: TimeValue | int | float):
        self.value %= other.value
        return self

# ----------------------------------------------------------------------------------------------------------------------
class Hour(TimeValue):
    def __repr__(self) -> str:
        return f"Hour(value={self.value})"
# ----------------------------------------------------------------------------------------------------------------------
class Minute(TimeValue):
    def __repr__(self) -> str:
        return f"Minute(value={self.value})"
# ----------------------------------------------------------------------------------------------------------------------
class Second(TimeValue):
    def __repr__(self) -> str:
        return f"Second(value={self.value})"
# ----------------------------------------------------------------------------------------------------------------------
class MilliSecond(TimeValue):
    def __repr__(self) -> str:
        return f"MilliSecond(value={self.value})"

